from typing import Optional, Union

import discord
from discord import PartialEmoji


class SelectOption:
    def __init__(
            self,
            label: str,
            value: Optional[str] = None,
            description: Optional[str] = None,
            emoji: Optional[Union[str, PartialEmoji]] = None,
            default: bool = False,
    ) -> None:
        self._label: str = label
        self._value: Optional[str] = label if value is None else value
        self._description: Optional[str] = description

        if isinstance(emoji, str):
            emoji = PartialEmoji.from_str(emoji)

        self._emoji: Optional[str] = emoji
        self._default: bool = default

    def label(self, label: str) -> 'SelectOption':
        self._label = label
        return self

    def value(self, value: str) -> 'SelectOption':
        self._value = value
        return self

    def description(self, description: str) -> 'SelectOption':
        self._value = description
        return self

    def emoji(self, emoji: Optional[Union[str, PartialEmoji]]) -> 'SelectOption':
        if isinstance(emoji, str):
            emoji = PartialEmoji.from_str(emoji)

        self._emoji = emoji
        return self

    def default(self, default: bool) -> 'SelectOption':
        self._default = default
        return self

    def to_discord_select_option(self) -> discord.SelectOption:
        return discord.SelectOption(
            label=self._label,
            value=self._value,
            description=self._description,
            emoji=self._emoji,
            default=self._default
        )



from typing import Optional, Tuple, Callable, Union

import discord
from discord import ui

from .component import Component
from .utils import _call_any


class MISSING:
    def __eq__(self, other):
        return False

    def __str__(self):
        return "MISSING"


class Message:
    def __init__(self,
                 content: Optional[str] = MISSING,
                 embed: Optional[discord.Embed] = None,
                 component: Optional[Component] = None) -> None:
        self.content: Optional[str] = content
        self.embed: Optional[discord.Embed] = embed
        self.component: Optional[Component] = component
        self.appear_func: Optional[Callable] = None
        self.disappear_func: Optional[Callable] = None

    async def update(self, other: 'Message') -> dict:
        kwargs = {}
        if self.content != other.content:
            kwargs['content'] = other.content
            self.content = other.content

        if self.embed != other.embed:
            kwargs['embed'] = other.embed
            self.embed = other.embed

        if self.component != other.component:
            kwargs['view'] = other.component.make_view() if other.component is not None else None
            self.component = other.component

        return kwargs

    def on_appear(self, func: Callable) -> 'Message':
        self.appear_func = func
        return self

    def on_disappear(self, func: Callable) -> 'Message':
        self.disappear_func = func
        return self

    async def appear(self) -> None:
        if self.appear_func is not None:
            await _call_any(self.appear_func)

    async def disappear(self) -> None:
        if self.disappear_func is not None:
            await _call_any(self.disappear_func)



from typing import Optional, Union, Callable

import discord
from discord import ui
from discord import ButtonStyle, PartialEmoji

from .item import Item
from .utils import _call_any


class CustomButton(ui.Button):
    def __init__(
            self,
            *,
            style: ButtonStyle,
            label: str,
            disabled: bool = False,
            custom_id: Optional[str] = None,
            url: Optional[str] = None,
            emoji: Optional[Union[str, PartialEmoji]] = None,
            row: Optional[int] = None,
            callback: Optional[Callable] = None,
    ) -> None:
        super(CustomButton, self).__init__(
            style=style,
            label=label,
            disabled=disabled,
            custom_id=custom_id,
            url=url,
            emoji=emoji,
            row=row
        )
        self.callback_func: Optional[Callable] = callback

    async def callback(self, interaction: discord.Interaction) -> None:
        if self.callback_func is None:
            return
        await _call_any(self.callback_func, interaction)


class Button(Item):
    def __init__(
            self,
            label: str = "",
            style: ButtonStyle = ButtonStyle.primary,
            disabled: bool = False,
            url: Optional[str] = None,
            emoji: Optional[Union[str, PartialEmoji]] = None,
            custom_id: Optional[str] = None,
    ) -> None:
        self._style: ButtonStyle = style
        self._label: str = label
        self._disabled: bool = disabled
        self._url: Optional[str] = url
        self._emoji: Optional[Union[str, PartialEmoji]] = emoji
        self._custom_id: Optional[str] = custom_id

        self._row: Optional[int] = None

        self.func: Optional[Callable] = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Button):
            return NotImplemented

        return self.to_dict() == other.to_dict()

    def to_discord(self) -> CustomButton:
        return CustomButton(
            style=self._style,
            label=self._label,
            disabled=self._disabled,
            url=self._url,
            emoji=self._emoji,
            row=self._row,
            callback=self.func,
            custom_id=self._custom_id
        )

    def to_dict(self) -> dict:
        return {
            'style': self._style,
            'label': self._label,
            'disabled': self._disabled,
            'emoji': self._emoji,
            'row': self._row,
            'callback': id(self.func),
            'custom_id': self._custom_id
        }

    def on_click(self, func: Callable) -> 'Button':
        self.func = func

        return self

    def style(self, style: discord.ButtonStyle) -> 'Button':
        self._style = style
        return self

    def label(self, label: str) -> 'Button':
        self._label = label
        return self

    def disabled(self, disabled: bool = False) -> 'Button':
        self._disabled = disabled
        return self

    def url(self, url: str) -> 'Button':
        self._url = url
        return self

    def emoji(self, emoji: Union[str, PartialEmoji]) -> 'Button':
        self._emoji = emoji
        return self

    def custom_id(self, custom_id: str) -> 'Button':
        self._custom_id = custom_id
        return self



from typing import Optional, List

from discord import ui

from .item import Item
from .select_option import SelectOption


class CustomSelect(ui.Select):
    pass


class Select(Item):
    def __init__(
            self,
            placeholder: Optional[str] = None,
            min_values: int = 1,
            max_values: int = 1,
            options: Optional[list] = None
    ) -> None:
        self._placeholder: Optional[str] = placeholder
        self._min_values: int = min_values
        self._max_values: int = max_values
        self._options: list = [] if options is None else options
        self._row: Optional[int] = None

    def placeholder(self, placeholder: str) -> 'Select':
        self._placeholder = placeholder
        return self

    def min_values(self, min_values: int) -> 'Select':
        self._min_values = min_values
        return self

    def max_values(self, max_values: int) -> 'Select':
        self._max_values = max_values
        return self

    def options(self, options: List[SelectOption]) -> 'Select':
        self._options = options
        return self

    def row(self, row: int) -> 'Select':
        self._row = row
        return self

    def to_discord(self) -> CustomSelect:
        return CustomSelect(
            placeholder=self._placeholder,
            min_values=self._min_values,
            max_values=self._max_values,
            options=[o.to_discord_select_option() for o in self._options],
            group=self._row
        )



# flake8: noqa
from .combine import ObservedObject, Published, State
from .button import Button
from .view import View
from .component import Component
from .message import Message
from .utils import async_interaction_partial, interaction_partial
from .select import Select
from .select_option import SelectOption


__version__ = "1.1.3"



from typing import Optional, List, Union

from discord import ui

from .item import Item


class Component:
    def __init__(self, items: Optional[List[Union[Item, List[Item]]]] = None) -> None:
        self.items: List[Union[Item, List[Item]]] = [] if items is None else items

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Component):
            return NotImplemented

        return self.items == other.items

    def make_view(self) -> ui.View:
        view = ui.View(None)
        i = 0
        for item in self.items:
            if not isinstance(item, list):
                view.add_item(item.to_discord())
                continue
            for item_ in item:  # type: Item
                setattr(item_, "_row", i)
                view.add_item(item_.to_discord())
            i += 1
        return view



import asyncio
from typing import Any, Callable

import discord


async def _call_any(func: Callable, *args: Any, **kwargs: Any) -> Any:
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    return func(*args, **kwargs)


def async_interaction_partial(func: Callable, *args: Any, **kwargs: Any) -> Callable:
    async def callback(interaction: discord.Interaction) -> Any:
        return await func(interaction, *args, **kwargs)
    return callback


def interaction_partial(func: Callable, *args: Any, **kwargs: Any) -> Callable:
    def callback(interaction: discord.Interaction) -> Any:
        return func(interaction, *args, **kwargs)

    return callback



from typing import Optional, List, Union, Callable, Any

import discord
from discord import ui
from discord.ext import commands

from .combine import State, ObservedObject
from .message import Message
from .view_manager import ViewManager


class View:
    def __init__(self, state: Union[discord.Client, commands.Bot]) -> None:
        self._watch_variables: List[str] = []

        self._state: Union[discord.Client, commands.Bot] = state
        self.manager = ViewManager()
        self.discord_message: Optional[discord.Message] = None
        self.view: Optional[ui.View] = None

        self._listeners: List[Callable] = []

    def add_listener(self, func: Callable, name: Optional[str] = None) -> None:
        if not isinstance(self._state, commands.Bot):
            raise ValueError("bot must be commands.Bot")

        self._state.add_listener(func, name)

    def remove_listener(self, func: Callable, name: Optional[str] = None) -> None:
        if not isinstance(self._state, commands.Bot):
            raise ValueError("bot must be commands.Bot")

        self._state.remove_listener(func, name)

    @staticmethod
    def listen(name: Optional[str] = None) -> Callable:
        def decorator(func: Callable) -> Callable:
            setattr(func, "__listener__", True)
            setattr(func, "__listener_name__", name or func.__name__)
            return func
        return decorator

    async def body(self) -> Message:
        return Message()

    def get_message(self) -> discord.Message:
        """
        Returns message that this View is attended
        :return: discord.Message
        """
        return self.manager.discord_message

    def _apply_listener(self):
        for name in dir(self):
            member = getattr(self, name, None)
            if hasattr(member, "__listener__"):
                self.add_listener(member, getattr(member, "__listener_name__", None))
                self._listeners.append(member)

    async def start(self, channel: discord.abc.Messageable) -> None:
        await self.manager.start(channel, await self.body())
        await self.manager.message.appear()
        self._apply_listener()

    async def apply(self, apply_message: discord.Message) -> None:
        await self.manager.apply(apply_message, await self.body())
        await self.manager.message.appear()
        self._apply_listener()

    async def stop(self) -> None:
        self.manager.raise_for_started()

        for listener in self._listeners:
            self.remove_listener(listener, getattr(listener, "__listener_name__", None))

        await self.manager.message.disappear()
        if self.view is not None:
            self.view.stop()

    async def update(self) -> None:
        self.manager.raise_for_started()
        await self.manager.update((await self.body()))

    def update_sync(self) -> None:
        if self._state is not None:
            self._state.loop.create_task(self.update())

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "_watch_variables":
            object.__setattr__(self, key, value)
            return

        if isinstance(value, ObservedObject):
            value.view = self

        if not hasattr(self, key):
            if isinstance(value, State):
                self._watch_variables.append(key)
                object.__setattr__(self, key, value.value)
                return
        if isinstance(value, State):
            value = value.value

        if key in self._watch_variables:
            object.__setattr__(self, key, value)
            self.update_sync()
            return

        object.__setattr__(self, key, value)



from setuptools import setup
import re

version = ''
with open('discord/ext/ui/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

if version.endswith(('a', 'b', 'rc')):
    # append version identifier based on commit count
    try:
        import subprocess

        p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += out.decode('utf-8').strip()
        p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += '+g' + out.decode('utf-8').strip()
    except Exception:
        pass

readme = ''
with open('README.rst') as f:
    readme = f.read()


setup(name='discord-ext-ui',
      author='sizumita',
      url='https://github.com/sizumita/discord-ext-ui',
      version=version,
      long_description=readme,
      long_description_content_type="text/x-rst",
      packages=['discord.ext.ui', 'discord.ext.ui.combine'],
      license='MIT',
      description='An Declarative UI Kit for discord.py',
      install_requires=[],
      python_requires='>=3.8.2',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: Japanese',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        ]
      )



from typing import Any, List, Optional, TYPE_CHECKING

from .published import Published

if TYPE_CHECKING:
    from ..view import View


class ObservedObject:
    def __init__(self) -> None:
        self._watch_variables: List[str] = []
        self.view: Optional['View'] = None

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "_watch_variables":
            object.__setattr__(self, key, value)
            return

        if not hasattr(self, key):
            if isinstance(value, Published) and hasattr(self, "_watch_variables"):
                self._watch_variables.append(key)
                object.__setattr__(self, key, value.value)
                return
        if isinstance(value, Published):
            value = value.value

        if key in getattr(self, "_watch_variables", []):
            object.__setattr__(self, key, value)
            if self.view is not None:
                self.view.update_sync()
            return

        return object.__setattr__(self, key, value)



from typing import Optional
import asyncio

import discord
from discord import ui

from .message import Message


class ViewManager:
    def __init__(self):
        self.message: Optional[Message] = None
        self.discord_message: Optional[discord.Message] = None
        self.view: Optional[ui.View] = None
        self.started: bool = False
        self.update_lock: asyncio.Lock = asyncio.Lock()

    def raise_for_started(self):
        if not self.started:
            raise Exception("View rendering is not started")

    async def send(self, channel: discord.abc.Messageable, message: Message) -> None:
        self.view = message.component.make_view()
        self.discord_message = await channel.send(
            content=message.content,
            embed=message.embed,
            view=self.view
        )

    async def start(self, channel: discord.abc.Messageable, message: Message) -> None:
        if self.started:
            raise ValueError("This View has already started")

        await self.send(channel, message)
        self.message = message
        self.started = True

    async def apply(self, apply_message: discord.Message, message: Message) -> None:
        if self.started:
            raise ValueError("This View has already started")

        self.view = message.component.make_view()
        self.discord_message = apply_message
        await self.discord_message.edit(
            content=message.content,
            embed=message.embed,
            view=self.view
        )
        self.message = message
        self.started = True

    async def update(self, message: Message):
        self.raise_for_started()
        async with self.update_lock:
            kwargs = await self.message.update(message)
            await self.render(**kwargs)

    async def render(self, **kwargs):
        await self.discord_message.edit(**kwargs)
        self.view = kwargs.get("view", None)
